/*
	Copyright 1985, 1986, 1987, 1988, 1989, 1990, 1991 Chris Lewis
		All Rights Reserved

    See the LICENSE file for a full description of restrictions under which
    this software is provided.

    Function:		Utility functions (font loads, prologs etc.)
 */

#include "defs.h"

#ifndef	lint
static char SCCSid[] =
    "@(#)utils.c: 2.14 Copyright 91/04/05 01:21:20 Chris Lewis";
#endif

#ifndef	HEADERSIZE
#	ifdef COFF
#		include <aouthdr.h>
#		include <filehdr.h>
#		include <scnhdr.h>
#		include <ldfcn.h>
#		include <reloc.h>
#		define H1 (sizeof (struct filehdr) + sizeof (struct aouthdr))
#		define H2 (sizeof (struct scnhdr) + sizeof (struct reloc))
#		define HEADERSIZE (H1 + H2)
#	else
#		include <a.out.h>
#		define HEADERSIZE (sizeof (struct exec))
#	endif
#endif

extern struct cattab tabN[], tabS[];

int lastFont, lastPoints;
int lastYPos, lastXPos;

int symidx = 3;		/* Default on otroff, otroff doesn't necessarily
			   emit loads either */

int extcount = 0;
struct troff2befont *extchars;
struct cattab *extidx;

extern char *malloc();

char *skipblanks(p)
register char *p; {
    while (*p && isspace(*p)) p++;
    return(p);
}

char *gettok(p, bp)
register char *p, *bp; {
    p = skipblanks(p);
    while(*p && !isspace(*p)) *bp++ = *p++;
    *bp = '\0';
    return(p);
}

interp(buf, xlator, suf)
char *buf, *suf; FUNC xlator; {
    register char *p;
    char token[512];
    p = gettok(buf, token);
    if (0 == strcmp(token, "include") || 0 == strcmp(token, "binclude")) {
	int binary;
	FILE *inc;
	binary = (token[0] == 'b') ? 1: 0;
	p = gettok(p, token);
	DBP((D_SPEC, "Trying to include %s\n", token));
	if (!(inc = fopen(token, "r"))) {
	    strcat(token, ".");
	    strcat(token, suf);
	    DBP((D_SPEC, "Trying to include %s\n", token));
	    if (!(inc = fopen(token, "r"))) {
		char nbuf[512];
		if (token[0] != '/') {
		    sprintf(nbuf, "%s/%s", LIBDIR, token);
		    DBP((D_SPEC, "Trying to include %s\n", nbuf));
		    inc = fopen(nbuf, "r");
		}
	    }
	}
	if (!inc) {
	    fprintf(stderr, "%s: cannot open file %s\n", progname, token);
	    exit(1);
	} else {
	    DBP((D_SPEC, "Successfull include\n"));
	    if (binary || !xlator)
		while ((binary = fread(token, 1, sizeof(token), inc)) > 0)
		    fwrite(token, 1, binary, stdout);
	    else
		(*xlator)(inc);
	    fclose(inc);
	}
    }

    else {
	fprintf(stderr, "%s: unknown directive: %s\n", progname, token);
	exit(1);
    }
}

char nodename[25];
getnodename() {
    FILE *popen();
    FILE *uuname;
    if ((uuname = popen(NODECMD, "r")) == NULL)
	strcpy(nodename, "<noname>");
    else {
	fscanf(uuname, "%s", nodename);
	if (pclose(uuname))
	    strcpy(nodename, "<noname>");
    }
}

struct fonttable fonttable[MAXFONTS+1];
struct fonttable *xlatetable[INTFONTS];

#define	READFONT	0
#define	READNORM	1
#define	READSYMB	2
#define	READEXT		3

loadfont(normfont, symfont)
struct troff2befont *normfont, *symfont; {

    char buffer[BUFSIZ];
    char *buf[6];
    int tableindex;
    int state, i, count;
    register char *first;
    extern int atoi();
    struct cattab *tab;
    struct troff2befont *otab;
    int done;

    FILE *fontfile;

    for (i = 0; i < 6; i++)
	buf[i] = mustmalloc(64, "loadfont");

    for (tableindex = 0; tableindex < INTFONTS; tableindex++) {
	xlatetable[tableindex] = &fonttable[tableindex];
    }

    fontfile = libopen(printer, "fonts");

    tableindex = 0;

    state = READFONT;

    while(fgets(buffer, sizeof(buffer), fontfile)) {

	first = buffer;

	for(;*first && isspace(*first); first++);

	if (!*first || *first == '#')
	    continue;

	switch(count = sscanf(first, "%s%s%s%s%s%s",
	    buf[0], buf[1], buf[2], buf[3], buf[4], buf[5])) {
	    case 1: buf[1][0] = '\0';
	    case 2: buf[2][0] = '\0';
	    case 3: buf[3][0] = '\0';
	    case 4: buf[4][0] = '\0';
	    case 5: buf[5][0] = '\0';
	    case 6:
		if        (strcmp(buf[0], "normal") == 0) {
		    state = READNORM;
		    break;
		} else if (strcmp(buf[0], "symbol") == 0) {
		    state = READSYMB;
		    break;
		} else if (strcmp(buf[0], "extensions") == 0) {
		    state = READEXT;
		    break;
		}
		switch(state) {
		    case READFONT:
			if (tableindex == MAXFONTS) {
			    fprintf(stderr, "%s: Too many fonts for %s!\n",
				progname, printer);
			    exit(1);
			}
			for(i = 0; i < 4; i++) {
			    first = mustmalloc(strlen(buf[i])+1,
				"fonttable");
			    strcpy(first, buf[i]);
			    fonttable[tableindex].tab[i] = first;
			}
#ifdef	OPT
			fonttable[tableindex].widthtable = (char *) NULL;
#endif
			tableindex++;
			break;
		    case READNORM: case READSYMB: case READEXT:
			done = 0;
			if (count <= 1) {
			    fprintf(stderr, "%s: bad %s.fonts line: %s\n",
				progname, printer, buffer);
			    exit(1);
			}

			if (state == READEXT) {
			    if (!(extcount % EXTCHUNK)) {
				if (!extchars) {
				    extchars = (struct troff2befont *)
					mustmalloc(EXTCHUNK *
					sizeof(struct troff2befont), "extchars");
				    extidx = (struct cattab *)
					mustmalloc((1 + EXTCHUNK) *
					sizeof(struct cattab), "extidx");
				} else {
				    extchars = (struct troff2befont *)
					realloc((char *) extchars,
					    (extcount + EXTCHUNK + 1) *
					    sizeof(struct troff2befont));
				    extidx = (struct cattab *)
					realloc((char *) extidx,
					    (extcount + EXTCHUNK + 1) *
					    sizeof(struct cattab));
				    clrarray(&extchars[extcount],
					EXTCHUNK * sizeof(struct troff2befont));
				    clrarray(&extidx[extcount],
					EXTCHUNK * sizeof(struct cattab));
				}
			    }

			    extidx[extcount].ch_name = mustmalloc(strlen(buf[0])
				+ 1, "extend");
			    strcpy(extidx[extcount].ch_name, buf[0]);

			    extidx[extcount].ch_catidx = extcount;
			    addchar(&extchars[extcount], buf, count, buffer);
			    extidx[extcount].ch_set = extchars[extcount].t2b_font;
			    extidx[extcount].ch_wididx = NTC;
			    extidx[extcount].ch_desc = "Ext";

			    extcount++;
			    extidx[extcount].ch_name = NOC;
			    done = 1;
			} else {
			    if (state == READNORM) {
				tab = tabN;
				otab = normfont;
			    } else {
				tab = tabS;
				otab = symfont;
			    }

			    for (i = 0; tab[i].ch_name; i++) {
				if (strcmp(tab[i].ch_name, buf[0]) == 0) {
				    addchar(&otab[i], buf, count, buffer);
				    done = 1;
				    break;
				}
			    }
			}
			if (!done) {
			    fprintf(stderr, "%s: couldn't apply %s.fonts: %s\n",
				progname, printer, buffer);
			    exit(1);
			}
		}
	}
    }
    fclose(fontfile);
#ifdef	OPT
    loadwidth(xlatetable[0]);
    loadwidth(xlatetable[1]);
    loadwidth(xlatetable[2]);
    loadwidth(xlatetable[3]);
#endif
    for (i = 0; i < 6; i++)
	free(buf[i]);
}

addchar(p, fontentry, count, buffer)
register struct troff2befont *p;
char *fontentry[];
char *buffer;
int count; {
    register char *first;
    switch(fontentry[1][0]) {
	case 'N':
	    p->t2b_font = N; break;
	case 'S':
	    p->t2b_font = S; break;
	case 'D':
	    p->t2b_font = D; break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    p->t2b_font = atoi(fontentry[1]);
	    break;
	default:
	    fprintf(stderr, "%s: bad font in %s\n",
		progname, buffer);
	    exit(1);
    }
    switch(count) {
	case 6:
	    p->t2b_scale = atoi(fontentry[5]);
	case 5:
	    p->t2b_yc = atoi(fontentry[4]);
	case 4:
	    p->t2b_xc = atoi(fontentry[3]);
	case 3:
	    octdecode(fontentry[2]);
	    first = mustmalloc(strlen(fontentry[2]) + 1,
		"fonttable");
	    strcpy(first, fontentry[2]);
	    p->t2b_charseq = first;
    }
}

/*	Translate any octal escapes *inplace* */
octdecode(str)
register char *str; {
    register char *from, *to;
    int tmp, i;
    from = to = str;
    while(*from) {
	if (*from != '\\')
	    *to++ = *from++;
	else {
	    tmp = 0;
	    from++;
	    i = 3;
	    while(i-- > 0 && *from && *from >= '0' && *from <= '7')
		tmp = (tmp << 3) + *from++ - '0';
	    *to++ = tmp&0xff;
	}
    }
    *to++ = '\0';
}


dumpsequence(sequence)
register char *sequence; {
    printf("\t");
    for(;*sequence;sequence++)
	if (isascii(*sequence) && isprint(*sequence))
	    printf("%c", *sequence);
	else
	    printf("\\%03o", (*sequence)&0xff);
}

dumpline(name, tp)
register char *name;
register struct troff2befont *tp; {
    if (name != NOC) {
	printf("%s\t", name);
	switch(tp->t2b_font) {
	    case N: printf("N"); break;
	    case S: printf("S"); break;
	    case D: printf("D"); break;
	    default: printf("%d", tp->t2b_font); break;
	}
	dumpsequence(tp->t2b_charseq);
	printf("\t%d\t%d\t%d", tp->t2b_xc, tp->t2b_yc, tp->t2b_scale);
	printf("\n");
    }
}

dumplist(namelist, tp)
register struct cattab *namelist;
register struct troff2befont *tp; {
    register int i;
    for (i = 0; namelist[i].ch_name && namelist[i].ch_catidx != NTC;
	i++, tp++) {
	if (strlen(namelist[i].ch_name) == 0)
	    continue;
	if (strcmp(namelist[i].ch_name, "hy") == 0)
	    dumpline("-", tp);
	dumpline(namelist[i].ch_name, tp);
    }
}

dumptables(bep)
struct backend *bep; {
    if (!bep) {
	fprintf(stderr, "%s: no backend selected for table dump\n", progname);
	exit(1);
    }
    printf("normal\n");
    dumplist(tabN, bep->bestdfont);
    printf("symbol\n");
    dumplist(tabS, bep->besymfont);
    if (extchars) {
	printf("extensions\n");
	dumplist(extidx, extchars);
    }
}

FontSel(from, to)
char from, *to; {
#ifdef	DEBUG
    register int i;
#endif
    register struct fonttable *p;
    int intfont;
    DBP((D_SPEC, "FontSel: %c -> %s\n", from, to));
    if (from < (ditroff? '0' : '1') || from > '9') {
	fprintf(stderr, "Bad arguments to FontSel: %c %s\n", from, to);
	return;
    }
    intfont = from - '1';
    if (intfont < 0)
	intfont = INTFONTS - 1;
    if (strcmp(to, "S") == 0) {
	symidx = intfont;
	DBP((D_CAT, "Special font is: %d\n", symidx));
    }

    for (p = fonttable; p->troffName; p++)
	if (strcmp(p->troffName, to) == 0) {
	    xlatetable[intfont] = p;
	    break;
	}
    if (!p->troffName) {
	fprintf(stderr, "Could not translate font %c (%s)\n", from, to);
    }
#ifdef	OPT
    if (p->troffName && !p->widthtable)
	loadwidth(p);
#endif

#ifdef	DEBUG
    for (i = 0; i < INTFONTS; i++)
	if (xlatetable[i])
	    DBP((D_SPEC, "Font %d->%s\n", i+1, xlatetable[i]->fontName));
#endif
}

FILE *libopen(pref, suf)
register char *pref, *suf; {

    register char *buffer = mustmalloc(strlen(LIBDIR) + 50, "libopen");
    register FILE *library;

    sprintf(buffer, "%s.%s", pref, suf);
    if ((library = fopen(buffer, "r")) == NULL) {
	sprintf(buffer, "lib/%s.%s", pref, suf);
	if ((library = fopen(buffer, "r")) == NULL) {
	    sprintf(buffer, "%s/%s.%s", LIBDIR, pref, suf);
	    if ((library = fopen(buffer, "r")) == NULL) {
		fprintf(stderr, "Cannot find %s.%s in . or lib or %s\n", pref, suf,
		    LIBDIR);
		exit(1);
	    }
	}
    }
    free(buffer);
    return(library);
}

resetState() {
    lastFont = -1;
    lastPoints = -1;
    lastYPos = -1;
    lastXPos = -1;
}

doprologs() {
    if (!prologs)
	return;
    while(*prologs)
	dospecial(*prologs++);
}

#ifdef	OPT
loadwidth(p)
struct fonttable *p; {
    FILE *f;
    int c;
    if ((int) p->widthtable == 1)
	return;
    p->widthtable = mustmalloc(224, "widthtable");
    strcpy(widthptr, "ft");
    strcat(widthptr, p->troffName);
    DBP((D_SPEC, "Attempting to read font file %s\n", widthtables));
    if (f = fopen(widthtables, "r")) {
	for (c = 0; c < HEADERSIZE; c++) getc(f);
#ifdef	ASCIIWIDTHS
	if (widread(p->widthtable, f)) {
#else
	if (224 != fread(p->widthtable, 1, 224, f)) {
#endif
	    fprintf(stderr, "%s: badly formed width table %s\n",
		progname, widthtables);
	    free(p->widthtable);
	    p->widthtable = (char *) 1;
	    DBP((D_SPEC, "Load of %s failed\n", widthtables));
	} else {
	    for (c = 0; c < 224; c++)
		p->widthtable[c] &= 0x3f;
	    DBP((D_SPEC, "Load of %s succeeded\n", widthtables));
	}
    } else {
	DBP((D_SPEC, "Failed to open widthtable %s\n", widthtables));
	free(p->widthtable);
	p->widthtable = (char *) 1;
    }
}

#ifdef	ASCIIWIDTHS
widread(widths, f)
char *widths;
FILE *f; {
    int v, i;
    for (i = 0; i < 224; i++)
	if (fscanf(f, "%d", &v) != 1)
	    return(1);
	else
	    widths[i] = v;
    return(0);
}
#endif

#endif

#ifdef	NULLCHECK
mab() {
    fprintf(stderr, "KABOOM\n");
    abort();
}
#endif

drawparse(ct, values, str)
int *ct;
short *values;
char *str; {
    int opcode, neg, temp;

    *ct = 0;
    while(*str && isspace(*str)) str++;
    opcode = *str;
    neg = 1;
    for (;*str;str++) {
	if (isspace(*str))
	    continue;
	else if (*str == '\\')
	    str += 3;
	else if (*str == '-')
	    neg = -1;
	else if (!isdigit(*str))
	    continue;
	else {
	    temp = *str - '0';
	    while(isdigit(*(str+1)))
		temp = temp * 10 + (*++str - '0');
	    values[(*ct)++] = neg * temp;
	    neg = 1;
	}
    }
    return(opcode);
}
